--[[
	author: Zax
	note: 
	content:
		globales: NPCM_
		functions: Npcm.
		event frame: npcm_frameMain
		client frame: npcm_frameClient
	todo:
--]]

---------------------------------------------------------------------------------------
------------------------------------- VARIABLES ---------------------------------------
local addonName = ...
local addonVar = Npcm.getVar()

-------------------------------------------- locales à ce fichier
local prevEvent = {["event"] = "", ["time"] = 0} -- pour délais event
local addonChan = "NPCMS_"..addonVar["compatVersion"] -- limite la communication aux versions compatibles de l'addon
----------- check et global check
local addonGChan = "NPCMS_GV" -- chan global, indépendant de la version
local prevCheck = {["players"] = {}, ["time"] = 0} -- pour comptage players check
local checkTimer
local checkDelay = 3 -- secondes
local numDaysCheckVers = 5 -- nombre de jours mini entre 2 check de version
local tblPlayers = {} -- pour check liste joueurs

function Npcm.getLocal(key)
	----------- pour traduction (voir localization.lua) : ne pas deplacer !!!
	if (NPCM_LOCAL_TBL[key] == nil) then return key else return NPCM_LOCAL_TBL[key] end
end;
-------------------------------------------- saved et init
NPCM_SAVED = {}
NPCM_SAVEDPERCHAR = {} -- pour log

local markersOrder = Npcm.getMarkersOrder()
local prevSavedMarkers = Npcm.initSavedMarkers() -- pour comparaison avec SAVED jouage de sons

function Npcm.upgradeVersion120() -- pour compatibilité NPCM_SAVED v >= 1.20
	if (NPCM_SAVED["settings"]["mj"]["hpprecision"] == nil) then
		NPCM_SAVED["settings"]["mj"]["hpprecision"] = 4
		for i, value in ipairs(NPCM_SAVED["markers"]) do
			NPCM_SAVED["markers"][i]["hp"] = 100 -- %
		end
		NPCM_SAVED["settings"]["mj"]["darkui"] = 1
		NPCM_SAVED["settings"]["mj"]["darkalpha"] = 0.8
	end
end

function Npcm.upgradeVersion130() -- pour compatibilité NPCM_SAVED v >= 1.30
	if (NPCM_SAVED["settings"]["minimap"] == nil) then
		NPCM_SAVED["settings"]["minimap"] = {}
		NPCM_SAVED["settings"]["minimap"]["visible"] = 1
		---------- pour correction bug HP
		for i, value in ipairs(NPCM_SAVED["markers"]) do
			NPCM_SAVED["markers"][i]["hp"] = 100 -- %
		end
		---------- supression unused
		NPCM_SAVED["settings"]["minimapIconAngle"] = nil
		NPCM_SAVED["settings"]["showMinimapIcon"] = nil
	end
end
	
---------------------------------------------------------------------------------------
----------------------------------------- EVENTS --------------------------------------

function Npcm.eventHandler(self,event,...)
	if (event == "PLAYER_LOGIN") then
		Npcm.chargement();
	elseif (event == "PLAYER_LOGOUT") then
		if (Npcm.getRank() == 0) then Npcm.videDataClient() end
		Npcm.saveSettings()
	elseif (event == "GROUP_ROSTER_UPDATE" or event == "RAID_ROSTER_UPDATE") then
		-- envoyé autant de fois que de membres !!!
		if (Npcm.waitEvent(event, 0.5)) then Npcm.rosterUpdate(event) end
	elseif (event == "PARTY_LEADER_CHANGED") then
		-- "GROUP_ROSTER_UPDATE" est envoyé juste avant !
		if (Npcm.waitEvent(event, 0.5)) then Npcm.leaderChanged(event) end
	elseif (event == "UNIT_SPELLCAST_SUCCEEDED") then -- placement de marker
		Npcm.spellOK(...) -- pour MJ
	elseif (event == "CHAT_MSG_ADDON") then
		Npcm.receivedMsg(...)
		
	elseif (event == "UPDATE_INSTANCE_INFO") then
		if (Npcm.inInstance() and Npcm.isActivated()) then -- second test pour event envoyé par journal aventurier
			Npcm.echoChatAppl(Npcm.getLocal("is not available in instance LFG/LFR."), false, "red")
			Npcm.activateClose()
		end
	end;
end

function Npcm.printEvent(event, isRaid, code, rank, comm, var) -- pour debug
	if (Npcm.debug()) then
		if (var == nil) then var = "" end
		if (comm == nil) then comm = "" else comm = "|cff00FF00"..comm.."|r" end
		local str = ".......... |cffFF8888"..code.."|r -----> "
		if (isRaid) then
			str = str.."RAID rank = "..rank.." ---> "..comm
		else
			str = str.."PARTY rank = "..rank.." ---> "..comm
		end
		if (var ~= "") then str = str.." ("..var..")" end
		print(str)
	end
end

function Npcm.rosterUpdate(event)
	if (Npcm.inInstance()) then
		Npcm.testPrint("|cffF46DF8"..event.."|r ------------------- inInstance : fermeture")
		Npcm.activateClose()
	else
		local rank = Npcm.getRank()
		if (Npcm.inRaid()) then -- attention à l'ordre du test : ne pas intervertir !
			------------------------------------------- vrai raid
			Npcm.testPrint("|cffF46DF8"..event.."|r ------------------- inRaid")
			
			if (rank == 2) then ---------------- MJ
				Npcm.printEvent(event, true, "RUMO", rank, "MJ - raid modified")
				if (npcm_frameMain:IsVisible()) then
					Npcm.setComm(Npcm.getLocal("Players changed."), 10)
					Npcm.sendUpdateClient(0, nil) -- le MJ envoit un update et force l'affichage des données à tous les membres
				end
				
			elseif (rank == 1) then ------------ nouveau RA
				Npcm.printEvent(event, true, "RURA", rank, "RA : nouveau RA") -- ne rien faire : voir LARA
			
			else ------------------------------- client
				if (npcm_frameMain:IsVisible()) then -- player (client) rejoint raid OU assistant est demote
					Npcm.printEvent(event, true, "RUDE", rank, "Joueur : assistant demote OU rejoignage raid ?")
					Npcm.activateClose() -- secu
				else
					Npcm.printEvent(event, true, "RUTU", rank, "Joueur : rejoignage Raid OU un joueur arrive ?") -- ne rien faire : voir LADE
				end
				Npcm.askUpdateClient() -- demande A au MJ (réponse U+S ou U+H ou I)
			end
		else
			------------------------------------------- groupe
			Npcm.testPrint("|cffF46DF8"..event.."|r ------------------- not inRaid (party)")
			if (rank >= 1) then
				Npcm.printEvent(event, false, "GUMO", rank, "MJ - group modified") -- ne rien faire : voir LORC
				Npcm.setComm(Npcm.getLocal("Players changed."), 10)
			elseif (rank == 0) then
				Npcm.printEvent(event, false, "GUNI", rank, "player - group modified") -- ne rien faire : voir LOZU
				--Npcm.askUpdateClient() -- demande A au MJ (réponse U+S ou U+H ou I)
			else
				--------------------------------------- ni raid ni groupe
				Npcm.printEvent(event, false, "GUQU", rank, "QUITTAGE raid ou groupe")
				Npcm.activateClose()
			end
		end
		Npcm.updatePartyIcon(rank)
	end
end

function Npcm.leaderChanged(event)
	C_Timer.After(1, function()
		if (Npcm.inInstance()) then
			Npcm.testPrint("|cffF46DF8"..event.."|r ------------------- inInstance : fermeture")
			Npcm.activateClose()
		else
			Npcm.updateMarkers() -- maj de SAVED
			local rank = Npcm.getRank()
			if (Npcm.inRaid()) then -- attention à l'ordre du test : ne pas intervertir !
				------------------------------------------- vrai raid
				Npcm.testPrint("|cffC570E4"..event.."|r ------------------- inRaid")
				if (rank == 2) then -- nouveau RL
					Npcm.printEvent(event, true, "LAMO", rank, "player new RL")
					if (Npcm.isActivated()) then
						npcm_frameClient:Hide()
						Npcm.askUpdateRA()
						Npcm.setComm(Npcm.getLocal("You are now DM."))
						--Npcm.activateMJ(true, "You are now DM.") -- true : send update
					end
				elseif (rank == 1) then -- nouveau RA
					Npcm.printEvent(event, true, "LARA", rank, "player new RA")
					if (Npcm.isActivated()) then
						npcm_frameClient:Hide()
						Npcm.activateMJ(false, Npcm.getLocal("You are now Assistant.")) -- false : NON send update
						Npcm.showHideAssistWarning(true)
						--Npcm.askUpdateClient()-- demande A au MJ (réponse U+S ou U+H ou B)
					end
				else -- client
					Npcm.printEvent(event, true, "LADE", rank, "Rejoignage raid ou changement de RL")
					--Npcm.askUpdateClient() -- demande A au MJ (réponse U+S ou U+H ou B)
				end
			else
				------------------------------------------- groupe
				Npcm.testPrint("|cffC570E4"..event.."|r ------------------- not inRaid (party)")
				if (rank >= 1) then
					Npcm.printEvent(event, false, "LORC", rank, "player devient leader")
					if (Npcm.isActivated()) then
						npcm_frameClient:Hide()
						Npcm.activateMJ(true, Npcm.getLocal("You are now DM.")) -- true : send update
					end
				elseif (rank == 0) then
					Npcm.printEvent(event, false, "LOZU", rank, "Rejoignage groupe ou leader groupe changed")
					npcm_frameMain:Hide()
					Npcm.askUpdateClient() -- demande A au MJ et réponse avec populate client et show UI client si l'UI MJ est visible
				else
					Npcm.printEvent(event, false, "LOQU", rank, "not grouped : QUITTAGE Party ?") -- non envoyé
				end
			end
			Npcm.updatePartyIcon(rank)
		end
	end)
end

function Npcm.waitEvent(event, delay) -- pour events envoyés à répétition : l'addon devient sourd pendant delay secondes
	if (delay == nil) then delay = 1; end -- 1 seconde
	if (event == prevEvent["event"]) then -- event reçu identique à celui stocké
		local timeDiff = time() - prevEvent["time"] -- secondes
		prevEvent["time"] = time()
		return (timeDiff >= delay) -- true si délai suffisant
	else
		prevEvent["event"] = event
		prevEvent["time"] = time()
		return true
	end
end

---------------------------------------------------------------------------------------
--------------------------------- CHARGEMENT & BINDINGS -------------------------------

function Npcm.chargement()
	------------------------------- 
	npcm_frameMain:RegisterEvent("PLAYER_LOGOUT")
	npcm_frameMain:RegisterEvent("CHAT_MSG_ADDON")
	npcm_frameMain:RegisterEvent("GROUP_ROSTER_UPDATE")
	npcm_frameMain:RegisterEvent("RAID_ROSTER_UPDATE")
	npcm_frameMain:RegisterEvent("PARTY_LEADER_CHANGED")
	-- seulement si UI ouverte (voir plus bas) :
	-- npcm_frameMain:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED"); -- UNIT_SPELLCAST_FAILED_QUIET		
	npcm_frameMain:RegisterEvent("UPDATE_INSTANCE_INFO")
	
	------------------------------- slash commands
	SlashCmdList["NPCM_CMD"] = Npcm.slashCmd;
		SLASH_NPCM_CMD1 = "/npcsmarkers";
		SLASH_NPCM_CMD2 = "/npcmarkers";
		SLASH_NPCM_CMD3 = "/npcm";
				
	------------------------------- bindings
	BINDING_HEADER_NPCM_TITLE = addonName
	BINDING_NAME_NPCMOPENCLOSE = Npcm.getLocal("Open / close NPCsMarkers")
	BINDING_NAME_NPCMUPDATE = Npcm.getLocal("Force data update")
	
	------------------------------- settings
	if (Npcm.isEmpty(NPCM_SAVED)) then
		NPCM_SAVED = Npcm.initData()
		if (Npcm.debug()) then print(addonVar["name"].." === no lua file : INIT SETTINGS ===") end
	else
		if (Npcm.isEmpty(NPCM_SAVED["settings"])) then
			NPCM_SAVED = Npcm.initData()
			if (Npcm.debug()) then print(addonVar["name"].." === empty settings : INIT SETTINGS ===") end
		else
			Npcm.upgradeVersion120()
			Npcm.upgradeVersion130()
			NPCM_SAVED["settings"]["mj"]["darkui"] = 1 -- DARKONLY
			local tbl = Npcm.initData()
			if (Npcm.tableKeyDiff(tbl, NPCM_SAVED, "") ~= nil) then
				NPCM_SAVED = tbl
				if (Npcm.debug()) then print(addonVar["name"].." === diff saved data INIT SETTINGS ===") end
			end
		end
	end
	--if (Npcm.debug()) then Npcm.echoTable(NPCM_SAVED["settings"], "NPCM_SAVED", false, "   "); end; --/////
	if (Npcm.debug()) then SetCVar("scriptErrors",1) end
	
	------------------------------- suite
	if (not Npcm.wowOk()) then
		Npcm.echoChatAppl(Npcm.getLocal("NOT loaded: not compatible with Classic environment."), true, "red")
	else
		if (Npcm.commPrefixOK(false)) then -- ouverture du chan de la version compatible de l'addon
			Npcm.minimapCreate(addonVar["minimap"], NPCM_SAVED["settings"]["minimap"]["posX"], NPCM_SAVED["settings"]["minimap"]["posY"]) -- fichier minimap.lua
			Npcm.minimapIconShow()
			Npcm.constructionFrames() -- frames hidden
			local str = Npcm.getLocal("loaded. Use the minimap button, or enter /npcsmarkers to display configuration interface.")
			if (Npcm.debug()) then str = str..Npcm.getTxtColor(" -- DEBUG ON.", "red") end
			Npcm.echoChatAppl(str, true, nil)
			Npcm.activateTest()
			Npcm.commPrefixOK(true) -- ouverture du chan global
			
			---------- demande de rafraichissement de l'UI au MJ si client (groupé)
			if (Npcm.getRank() == 0) then Npcm.askUpdateClient() end -- client d'un party ou raid (mais pas instance)
			---------- settings panel (voir fichier _settings.lua)
			if (not Npcm.isEmpty(addonVar["settingsPanel"])) then Npcm.settingsPanelUpdate() end
		else
			Npcm.activateClose()
			Npcm.echoChatAppl(Npcm.getLocal("NOT loaded due to addon communication prefix error."), true, "red")
		end
	end
end

---------------------------------------------------------------------------------------
-------------------------------------- SLASH CMD --------------------------------------
function Npcm.slashCmd(arg, execLevel)
	Npcm.slashCmd_Lib(arg, execLevel)
end

function Npcm.slashCmd_getList(arg2) -- liste des commandes-slash (voir .lib)
	-- .slashCmd_showList(dontPrint) -- print la liste en print, ou la renvoit en chaine avec dontPrint à true (pour frame cfg)
	local tbl = {
		["macro"] = {
			{
				["cmd"] = {"f_cleanline"},
				["func"] = function() Npcm.cleanMarker(Npcm.entierParDefaut(arg2, 1)) end,
				["label"] = "INTERNAL: pour macro text btn secure",
			},
			{
				["cmd"] = {"f_typemarker"}, -- pour btn secure MJ
				["func"] = function() Npcm.chatEditInsertMarkerMJ(arg2) end,
				["label"] = "INTERNAL: pour macro text btn secure",
			},
		},
		
		["user"] = {
			{
				["cmd"] = {"settings"},
				["type"] = "",
				["func"] = function() Npcm.printAddonSettings() end,
				["label"] = Npcm.getLocal("display actual settings"),
			},
			{
				["cmd"] = {"show", "open", "off"},
				["type"] = "",
				["func"] = function() Npcm.activateOpen() end,
				["label"] = Npcm.getLocal("open/close addon"),
			},
			{
				["cmd"] = {"hide", "close", "off"},
				["type"] = "",
				["func"] = function() Npcm.activateClose() end,
				["label"] = Npcm.getLocal("open/close addon"),
			},
			{
				["cmd"] = {"expand"},
				["type"] = "yes | no",
				["func"] = function() Npcm.expandUI(arg2) end,
				["label"] = Npcm.getLocal("expand/collapse addon UI"),
			},
			{
				["cmd"] = {"reset"},
				["func"] = function() Npcm.reset() end,
				["label"] = Npcm.getLocal("reset addon UI position and scale"),
			},
			{
				["cmd"] = {"check"},
				["func"] = function() Npcm.commCheck() end,
				["label"] = Npcm.getLocal("perform version and communication check"),
			},
			{
				["cmd"] = {"refresh"},
				["func"] = function() Npcm.refresh() end,
				["label"] = Npcm.getLocal("refresh markers data"),
			},
			
			{
				["cmd"] = {"help", "aide", "?"},
				["func"] = function() Npcm.help() end,
				["label"] = "",
			},
		},
		
		["dev"] = {
			{
				["cmd"] = {"test"},
				["func"] = function() Npcm.fcnTest(arg2) end,
				["label"] = "DEV test with 1 argument",
			},
			{
				["cmd"] = {"debug"},
				["func"] = function() Npcm.toggleDebug() end,
				["label"] = "toggle debug",
			},
			{
				["cmd"] = {"client"},
				["func"] = function()
					if (Npcm.debug()) then
						Npcm.fcnForceClient();
					else
						Npcm.echoChatAppl("ERROR: testing mode required for '"..arg2.."'.", false, "red")
					end;
				end,
				["label"] = "force client mode",
			},
			{
				["cmd"] = {"mjdata"},
				["func"] = function() Npcm.printMarkersData() end,
				["label"] = "print MJ markers data",
			},
			{
				["cmd"] = {"markers"}, -- print les markers visibles
				["func"] = function()
					Npcm.showRealMarkers()
				end,
				["label"] = "print placed markers",
			},
			{
				["cmd"] = {"pop", "populate", "peuple"},
				["func"] = function()
					if (Npcm.debug()) then
						Npcm.popDev() -- pour dvlpt
					else
						Npcm.echoChatAppl("ERROR: testing mode required for '"..arg2.."'.", false, "red")
					end
				end,
				["label"] = "populate test data",
			},
			{
				["cmd"] = {"scale"},
				["func"] = function() Npcm.scale(Npcm.videSiNul(arg2)) end,
				["label"] = "scale addon UI",
			},
			{
				["cmd"] = {"alpha"},
				["func"] = function() Npcm.setAlpha(Npcm.videSiNul(arg2)) end,
				["label"] = "set addon UI alpha",
			},
		},
		["DEFAULTCMD"] = "show", -- commande par défaut
		["DEFAULTARG"] = "help", -- argument par défaut d'une commande
	}
	return tbl
end

function Npcm.activate(booleen) -- MJ et client : reg events et show bon frame
	local val
	if (booleen == nil) then
		----------- switch
		val = (not Npcm.isActivated())
	else
		----------- given value
 		val = Npcm.isTrue(booleen, true)
	end
	if (val == true) then Npcm.activateOpen() else Npcm.activateClose(true) end
end

function Npcm.activateOpen()
	if (not Npcm.inInstance()) then
		npcm_frameClient:Hide()
		npcm_frameMain:Hide()
		local rank = Npcm.getRank()
		Npcm.testPrint("activateOpen() : rank = "..rank)
		if (rank >= 1) then -- RL, RA ou chef groupe
			npcm_frameMain:Show()
			Npcm.activateMJ(rank == 2) -- updateMarkers + populateFrameMJ (+ sendUpdateClient() si RL)
			Npcm.updatePartyIcon(rank)
		elseif (rank == 0) then -- joueur
			prevSavedMarkers = Npcm.initSavedMarkers()
			Npcm.askUpdateClient()-- demande d'update au MJ (et réponse du MJ si son UI est visible)
			--Npcm.echoChatAppl("is now active.", false, "blue")
		else -- non groupé
			npcm_frameMain:Show()
			Npcm.updateMarkers()
			Npcm.populateFrameMJ(Npcm.getLocal("Not in home party or raid."), 10)
		end
		
		NPCM_SAVED["settings"]["visible"] = 1
		Npcm.minimapOnOff(true)

		if (Npcm.isTrue(NPCM_SAVED["settings"]["forceClient"])) then npcm_frameClient:Show() end -- affichage simple (sans update) pour dev
	else
		Npcm.echoChatAppl(Npcm.getLocal("is not available in instance LFG/LFR."), false, "red")
	end
end

function Npcm.activateClose(userActionBoolean)
	if (Npcm.isEmpty(userActionBoolean)) then userActionBoolean = false end
	
	if (Npcm.isActivated()) then
		if (Npcm.getRank() == 2) then Npcm.sendVideClient() end -- seulement le vrai MJ
		Npcm.showHideFrame(false)
		if (not userActionBoolean) then NPCM_SAVED["settings"]["visible"] = 0 end
	end
	npcm_playersFrame:Hide() -- secu
	Npcm.minimapOnOff(false) -- secu
	npcm_frameMain:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED") -- secu
end

function Npcm.activateMJ(sendClientBoolean, comm)
	if (comm == nil) then comm = Npcm.getLocal("Ready.") end --accepte "" pour effacer le comm
	npcm_frameMain:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED") -- secu
	if (Npcm.updateMarkers()) then -- maj de SAVED
		Npcm.populateFrameMJ(comm, 10)
		npcm_frameMain:Show()
		if (sendClientBoolean) then Npcm.sendUpdateClient(0, comm) end -- envoi au client cmd update toutes les lignes
	else
		Npcm.echoChatAppl(Npcm.getLocal("Unexpected markers error."), false, "red")
	end
end
		
function Npcm.activateTest()
	if (Npcm.debug()) then
		if (Npcm.hasMarkers()) then
			if (Npcm.getRank() >= 1) then Npcm.activateOpen() end -- RL, RA ou chef groupe
		end
	end
end

function Npcm.printAddonSettings()
	Npcm.echoChatAppl("settings", true, "grey")
	for key, value in pairs(NPCM_SAVED["settings"]) do
		if (type(value) == "table") then
			Npcm.echoChat("    "..key..":", "SYSTEM", true)
			for subkey, subvalue in pairs(value) do
				if (type(subvalue) == "table") then value = "TABLE" end
				Npcm.echoChat("        "..subkey..": "..subvalue, "SYSTEM", true)
			end
		else
			Npcm.echoChat("    "..tostring(key)..": "..tostring(value), "SYSTEM", true)
		end
	end
end

function Npcm.reset()
	wipe(NPCM_SAVED)
	NPCM_SAVED = Npcm.initData()
	npcm_frameMain:SetScale(Npcm.nombreParDefaut(NPCM_SAVED["settings"]["mj"]["scale"], 1))
	--npcm_frameMain:SetAlpha(1)-- pas toucher à la barre de titre
	npcm_frameBk:SetAlpha(Npcm.nombreParDefaut(NPCM_SAVED["settings"]["mj"]["alpha"], 1))
	npcm_frameBk:SetAlpha(Npcm.nombreParDefaut(NPCM_SAVED["settings"]["mj"]["darkalpha"], 1))
	--CLIPREFS
	
	Npcm.populateFrames()
	Npcm.activateOpen()
	Npcm.expandUI(1)
	if (Npcm.debug()) then
		Npcm.printAddonSettings()
	else
		Npcm.echoChatAppl(Npcm.getLocal("settings reset to default."), "false", "");
	end
end

function Npcm.commCheck()
	if (not npcm_playersFrame_mask:IsVisible()) then -- ne pas renvoyer la cmd si check en cours
		if (Npcm.sendCheck()) then -- envoi de "C"
			--Npcm.echoChatAppl("Checking players. Wait a few seconds...", false, "blue")
			tblPlayers = Npcm.commCheckGetList()
			Npcm.listPlayersShow()
			wipe(prevCheck)
			prevCheck = {["players"] = {}, ["time"] = time()} -- secondes sur Windows
			
			
			if (checkTimer ~= nil) then checkTimer:Cancel() end -- clean et init
			checkTimer = C_Timer.NewTimer(7, function() -- en cas de non-réponse (membre déco par exemple)
				--Npcm.echoTable(checkTimer,"checkTimer FIN")
				Npcm.testPrint("commCheck() : TIMEOUT -----------")
				table.sort(tblPlayers, function(a,b) return a.sorting < b.sorting end)
				Npcm.listScrollerUpdate()
				Npcm.listPlayersOK()
			end)
			-- suite sur Npcm.logMembers()
		else
			Npcm.echoChatAppl(Npcm.getLocal("Unable to communicate."), false, "red")
		end
	end
end

function Npcm.commCheckGetList() -- liste des joueurs party ou raid (doit etre indexée pour FauxScroll, et non associative !)
	local playerName, serverName = UnitFullName("player") -- nom condensé du server (sans espace)
	
	wipe(tblPlayers)
	
	local vers, active, unitType, maxItem
	if (Npcm.inRaid()) then -- attention à l'ordre du test : ne pas intervertir !
		unitType = "raid"
		maxItem = GetNumGroupMembers()
	elseif (Npcm.isGrouped(false)) then
		unitType = "party"
		maxItem = 4
	end
		
	for i = 1, maxItem do
		unit = unitType..i
		if (UnitExists(unit)) then
			name, realm = UnitFullName(unit)
			if (not realm) then realm = serverName end
			if (name ~= playerName) then
				tblPlayers[i] = {["fullname"] = name.."-"..realm, ["name"] = name, ["active"] = 0, ["version"] = "", ["status"] = 0, ["sorting"] = "0_"..name}
			else
				tblPlayers[i] = {["fullname"] = name.."-"..realm, ["name"] = "* "..name, ["active"] = 0, ["version"] = "", ["status"] = 0, ["sorting"] = "0_*"..name}
			end
		end
	end
	if (unitType == "party") then -- ajout soi-même
		table.insert(tblPlayers, {["fullname"] = playerName.."-"..serverName, ["name"] = "* "..playerName, ["active"] = Npcm.booleanToInt(Npcm.isActivated()), ["version"] = Npcm.getVersionStr(), ["status"] = 0, ["sorting"] = "0_*"..playerName})
	end 

	return tblPlayers
end

function Npcm.listGetPlayers() -- MAJ de la liste des joueurs party ou raid avec versions
--Npcm.echoTable(tblPlayers, "tblPlayers")--/////
	return tblPlayers -- pour UI
end

function Npcm.commCheckUpdateList(senderName, version, active, rank) -- senderName est le fullName d'un joueur (avec nom serveur)
	-- rank non géré dans tblPlayers
	for i, value in ipairs(tblPlayers) do
		if (value["fullname"] == senderName) then
			tblPlayers[i]["active"] = active
			tblPlayers[i]["version"] = version
			if (Npcm.zeroSiNul(version, true) < tonumber(addonVar["version"])) then
				tblPlayers[i]["status"] = 1
				tblPlayers[i]["sorting"] = "1_"..value["name"]
			else
				tblPlayers[i]["status"] = 2
				tblPlayers[i]["sorting"] = "2_"..value["name"]
			end
			break
		end
	end
	if (npcm_playersFrame:IsVisible()) then Npcm.listScrollerUpdate() end -- pour effet visuel
end

function Npcm.logMembers(senderName, version, active, rank) -- "R" reçu par Npcm.answerCheck()
	if (checkTimer ~= nil) then checkTimer:Cancel() end -- re-init du timer à chaque réponse
	--[[
		retour v < 1.03 : pas de params 'active' et 'rank' dans retour 'R' et ordre inversé
		Npcm.logMembers(senderName, version, active)
	--]]
	if (active == nil) then active = "?" end -- retour v < 1.03
	if (rank == nil) then rank = "?" end -- retour v < 1.03
	Npcm.testPrint("logMembers() : player response : "..senderName.." (version="..Npcm.nilSiNul(version).." active="..active.." rank="..rank..") --> checkTimer augmented.")
	Npcm.commCheckUpdateList(senderName, Npcm.zeroSiNul(version), active, rank)
	------------- fin
	checkTimer = C_Timer.NewTimer(checkDelay, function()
		if (npcm_playersFrame:IsVisible()) then
			table.sort(tblPlayers, function(a,b) return a.sorting < b.sorting end)
			Npcm.listScrollerUpdate()
			Npcm.listPlayersOK()
		else -- OLD sans UI list scroll
			local group = "party"
			if (Npcm.inRaid()) then group = "raid" end
			Npcm.echoChatAppl("Communication checked : "..Npcm.zeroSiNul(#prevCheck["players"], true).." other player(s) found in "..group..".", true, "blue")
		end
	end)
		
	if (not tContains(prevCheck["players"], senderName)) then
		table.insert(prevCheck["players"], Npcm.getLogMember(senderName, Npcm.zeroSiNul(version),  Npcm.zeroSiNul(active)))
	end
end

function Npcm.getLogMember(name, version, active)
	return ({["name"] = name, ["version"] = version, ["active"] = active})
end

function Npcm.refresh()
	if (Npcm.isMJ()) then
		------- btn Update
		if (Npcm.updateMarkers()) then
			Npcm.populateFrameMJ(nil)
			Npcm.sendUpdateClient(0, nil)
		else
			Npcm.echoChatAppl(Npcm.getLocal("Unexpected markers error."), false, "red")
		end
	else
		------- client
		Npcm.askUpdateClient() -- demande d'update au MJ
	end
end

function Npcm.scale(value, noAlertBoolean)
	local scale = tonumber(value)
	if (scale and scale >= 0.5 and scale <= 1) then
		npcm_frameMain:SetScale(scale)
		NPCM_SAVED["settings"]["mj"]["scale"] = scale
	else
		if (noAlertBoolean ~= true) then Npcm.echoChatAppl(Npcm.getLocal("Scale").." "..Npcm.getLocal("must be between 0.5 and 1."), false, "red") end
	end
end

function Npcm.setAlpha(value, noAlertBoolean)
	local alpha = tonumber(value)
	if (alpha and alpha >= 0.5 and alpha <= 1) then
		npcm_frameBk:SetAlpha(alpha)
		if (NPCM_SAVED["settings"]["mj"]["darkui"] == 1) then
			NPCM_SAVED["settings"]["mj"]["darkalpha"] = alpha
		else
			NPCM_SAVED["settings"]["mj"]["alpha"] = alpha
		end
	else
		if (noAlertBoolean ~= true) then Npcm.echoChatAppl(Npcm.getLocal("Alpha").." "..Npcm.getLocal("must be between 0.5 and 1."), false, "red") end
	end
end

function Npcm.help()
	Npcm.showConfig() -- pour compatibilité
end

function Npcm.showConfig()
	if (Settings ~= nil) then -- wow10
		local settingsCategoryID = _G[addonVar["settingsPanel"]].categoryID -- voir fichier _settings.lua
		Settings.OpenToCategory(settingsCategoryID) -- ouverture option panel directe
	else -- classic
		InterfaceAddOnsList_Update()
		InterfaceOptionsFrame_OpenToCategory(addonName) -- wow9
	end
end

function Npcm.popDev() -- pour dev
	--------------------------------- markers
	if (npcm_frameMain:IsVisible()) then
		wipe(NPCM_SAVED["markers"]);
		NPCM_SAVED["markers"] = Npcm.initSavedMarkers();
		NPCM_SAVED["markers"][1] = {["text"] = "3 grunts with axes", ["hp"] = 4, ["boss"] = 0, ["attitude"] = 0, ["visible"] = 0};
		NPCM_SAVED["markers"][2] = {["text"] = "Warchief", ["hp"] = 2, ["boss"] = 1, ["attitude"] = 0, ["visible"] = 0};
		NPCM_SAVED["markers"][3] = {["text"] = "Ancient kaldorei altar", ["hp"] = 4, ["boss"] = 0, ["attitude"] = 1, ["visible"] = 0};
		NPCM_SAVED["markers"][4] = {["text"] = "Headhunter Troll", ["hp"] = 3, ["boss"] = 0, ["attitude"] = 0, ["visible"] = 0};
		
		Npcm.populateFrames("Pop executed.", 10);
	else --if (npcm_playersFrame:IsVisible()) then
		--------------------------------- liste players
		wipe(tblPlayers)
		tblPlayers[1] = {["fullname"] = "Leeroy-KirinTor", ["name"] = "Leeroy", ["active"] = 0, ["version"] = "", ["status"] = 0, ["sorting"] = "0_Leeroy"}
		tblPlayers[2] = {["fullname"] = "Arjuna-KirinTor", ["name"] = "Arjuna", ["active"] = 1, ["version"] = "1.01", ["status"] = 2, ["sorting"] = "2_Arjuna"}
		tblPlayers[3] = {["fullname"] = "Fafner-KirinTor", ["name"] = "Fafner", ["active"] = 1, ["version"] = "1.03", ["status"] = 2, ["sorting"] = "2_Fafner"}
		tblPlayers[4] = {["fullname"] = "Nina-KirinTor", ["name"] = "* Nina", ["active"] = 1, ["version"] = "1.03", ["status"] = 2, ["sorting"] = "2_Nina"}	
		tblPlayers[5] = {["fullname"] = "Hichinji-KirinTor", ["name"] = "Hichinji", ["active"] = 0, ["version"] = "1.03", ["status"] = 1, ["sorting"] = "1_Hichinji"}
		--[[
		local vers = Npcm.getVersionStr()
		for i = 1, 35 do --35
			tblPlayers[i] = {["fullname"] = "Player"..i.."-KirinTor", ["name"] = "Player"..i, ["active"] = math.random(1, 2) - 1, ["version"] = vers, ["status"] = math.random(1, 3) - 1, ["sorting"] = ""}
			tblPlayers[i]["sorting"] = tblPlayers[i]["status"].."_"..tblPlayers[i]["name"]
		end
		--]]
		table.sort(tblPlayers, function(a,b) return a.sorting < b.sorting end)
		Npcm.listPlayersShow()
		Npcm.listScrollerUpdate()
		Npcm.listPlayersOK()
	end
end

function Npcm.showRealMarkers() -- debug
	print("------------ visible markers -----------")
	for i = #markersOrder, 1, -1 do
		if (IsRaidMarkerActive(i)) then -- i : ordre des markers !
			print(i.."    ".._G["WORLD_MARKER"..i])
		end
	end
	print("---------------------------------------------")
end

function Npcm.fcnTest(arg) --$$$
	--Npcm.printFramePos("npcm_frameClient")
	--print("------------------------------------------")
	print(date("%y.%m.%d %H:%M:%S", GetServerTime()))
	local version, build, datevers, tocversion = GetBuildInfo()
	print(version, build, datevers, "toc = "..tocversion)
	--print("------------------------------------------")
	--Npcm.echoGroupStatus()
	--print("ActionButtonUseKeyDown CVAR=", GetCVar("ActionButtonUseKeyDown"))
	--SetCVar("ActionButtonUseKeyDown", 1)
	
	Npcm.printMacroTxts()
end

function Npcm.printMacroTxts() -- DEV
	if (Npcm.isMJ()) then
		print("------------- MJ markers ------------")
		local f, nom, txt
		local cpt = 0
		for i, value in ipairs(NPCM_SAVED["markers"]) do
			if (Npcm.isTrue(value["visible"])) then
				cpt = cpt + 1
				nom = _G["npcm_frameBk_fLine_"..i.."_editBox"]:GetText()
				f = _G["npcm_frameBk_fLine_"..i.."_icn"]
				print(i, "---#", cpt, "---nom=", nom, "---sMacro1=", Npcm.nilSiNul(f:GetAttribute("shift-macrotext1")))
			end
		end
	else
		print("------------- Players markers ------------")
		local f, desc, numIcn
		local cpt = 0
		for i, value in pairs(NPCM_SAVED["markers"]) do
			if (Npcm.isTrue(value["visible"])) then
				cpt = cpt + 1
				f = _G["npcm_frameClient_fLine_"..cpt]
				print(i, "---#", cpt, "---desc=", f.desc, "---numIcn=", f.numIcn)
			end
		end
	end
end

function Npcm.showChannels()
	for i = 1, MAX_WOW_CHAT_CHANNELS do
		local channelID, channelName = GetChannelName(i);
		if (channelName) then
			print("CHANNEL "..i.." - "..channelName)
		end
	end
end

function Npcm.toggleDebug()
	if (Npcm.isTrue(NPCM_SAVED["settings"]["debug"])) then
		NPCM_SAVED["settings"]["debug"] = 0
	else
		NPCM_SAVED["settings"]["debug"] = 1
	end
	Npcm.echoChatAppl("debug "..tostring(Npcm.debug())..".", true, "orange")
end

function Npcm.fcnForceClient()
	NPCM_SAVED["settings"]["forceClient"] = 1 - tonumber(NPCM_SAVED["settings"]["forceClient"])
	Npcm.echo(addonVar["name"]..": forcage client "..tostring(NPCM_SAVED["settings"]["forceClient"]).." taper /npcm pour recharger.", "orange");
end

---------------------------------------------------------------------------------------
---------------------------------------- INTERNE --------------------------------------

function Npcm.isActivated()
	return (npcm_frameMain:IsVisible() or npcm_frameClient:IsVisible())
end

function Npcm.showHideFrame(booleen) -- false pour masquer tous les frames, et true pour n'afficher que le bon (MJ ou client)
	npcm_frameClient:Hide()
	npcm_frameMain:Hide()
	if (booleen) then
		if (Npcm.isMJ()) then
			npcm_frameMain:Show()
		else
			npcm_frameClient:Show();
		end
	end
end

function Npcm.saveSettings()
	-- les autres settings sont sauvegardés après chaque modif user ou sur la validation de l'UI config
	Npcm.saveFramesPosition()
end

function Npcm.saveFramesPosition()
	NPCM_SAVED["settings"]["mj"]["posX"] = floor(tonumber(npcm_frameMain:GetLeft())) -- distance Left / coin BG écran
	NPCM_SAVED["settings"]["mj"]["posY"] = floor(tonumber(npcm_frameMain:GetTop())) -- distance Haut / coin BG écran
	NPCM_SAVED["settings"]["mj"]["scale"] = npcm_frameMain:GetScale()
	NPCM_SAVED["settings"]["client"]["posX"] = floor(tonumber(npcm_frameClient:GetLeft())) -- distance Left / coin BG écran
	NPCM_SAVED["settings"]["client"]["posY"] = floor(tonumber(npcm_frameClient:GetTop())) -- distance Haut / coin BG écran
	
	NPCM_SAVED["settings"]["minimap"]["posX"] = floor(tonumber(_G[addonVar["minimap"]]:GetLeft()))
	NPCM_SAVED["settings"]["minimap"]["posY"] = floor(tonumber(_G[addonVar["minimap"]]:GetTop()))
end

function Npcm.isMJ() -- fonction générale : accepte les groupes mais pas les instances
	-- renvoit true si l'UI du MJ peut être affichée
	if (Npcm.isTrue(NPCM_SAVED["settings"]["forceClient"])) then
		return false
	else
		if (Npcm.inRaid()) then -- attention à l'ordre du test : ne pas intervertir !
			return (Npcm.getRaidRank() >= 1)
		elseif (Npcm.inParty()) then
			return UnitIsGroupLeader("player")
		else
			return false -- non groupé ou instance
		end
	end
end

function Npcm.isRaidLeader() -- fonction stricte pour raid
	return (Npcm.getRaidRank() == 2)
end

function Npcm.isRaidAssist() -- fonction stricte pour raid
	return (Npcm.getRaidRank() == 1)
end

function Npcm.getRank()
	-- renvoit 2 si RL ou groupe leader si non raid
	-- 1 si RA
	-- 0 si client
	-- -1 si non groupé
	-- -2 si en instance
	if (Npcm.inInstance()) then
		return -2
	elseif (Npcm.inRaid()) then
		return Npcm.getRaidRank()
	else
		if (GetNumGroupMembers() > 0) then
			if (UnitIsGroupLeader("player")) then return 2 else return 0 end
		else
			return -1
		end
	end
end

function Npcm.getRaidRank() -- renvoit le rank dans le raid : 2 pour RL, et 1 pour RA
	if (Npcm.inRaid()) then
		local playerName = UnitName("player")
		for i = 1, MAX_RAID_MEMBERS do
			local name, rank = GetRaidRosterInfo(i)
			if (name == playerName) then return rank end
		end
	end
	return 0
end

function Npcm.echoGroupStatus() -- pour debug
	Npcm.echoChatAppl("DEV: party status:", false, "orange")
	Npcm.echoDebug(Npcm.isGrouped(), "Npcm.isGrouped() -- true if raid or party")
	Npcm.echoDebug(Npcm.inParty(), "Npcm.inParty() -- party only")
	Npcm.echoDebug(Npcm.inRaid(), "Npcm.inRaid() -- raid only")
	Npcm.echoDebug(Npcm.inInstance(), "Npcm.inInstance() -- instance only")
	
	Npcm.echoDebug(Npcm.getRank(), "Npcm.getRank() -- 2=raid/party leader | 1=RA | 0=player | -1=not in party")
	Npcm.echoDebug(Npcm.isMJ(), "Npcm.isMJ() -- true if leader/RA")
	Npcm.echoDebug(Npcm.isRaidLeader(), "Npcm.isRaidLeader() -- raid only")
	Npcm.echoDebug(Npcm.isRaidAssist(), "Npcm.isRaidAssist() -- raid only")
	Npcm.echoDebug(Npcm.getRaidRank(), "Npcm.getRaidRank() -- raid only")
	
	Npcm.echoDebug(GetNumGroupMembers(), "GetNumGroupMembers()")
	Npcm.echoDebug(IsInGroup(LE_PARTY_CATEGORY_HOME), "IsInGroup(HOME) -- party only")
	Npcm.echoDebug(IsInRaid(), "IsInRaid() -- raid or instance")
	Npcm.echoDebug(IsInInstance(), "IsInInstance()")
	Npcm.echoDebug(IsInGroup(LE_PARTY_CATEGORY_INSTANCE ), "IsInGroup(INSTANCE ) -- instance only")

	Npcm.echo("---------------------------", "orange")
end

function Npcm.frameExists(frame)
	return (frame ~= nil)
end

function Npcm.hasMarkers() -- true si des markers ont été posés
	for i = 1, #markersOrder do
		if (IsRaidMarkerActive(i) == true) then return true end
	end
	return false
end

function Npcm.getPlayerFullName()
	local playerName, realm = UnitFullName("player")
	return (Npcm.nilSiNul(playerName).."-"..Npcm.nilSiNul(realm))
end

function Npcm.wowOk()
	local retOK, ret1 = pcall(IsRaidMarkerActive, 1)
	return (retOK == true) -- forcer le false
end

function Npcm.escapeQuote(str)
	if (str ~= nil) then
		local output = string.gsub(str, "\"", "'")
		return output
	else
		return str
	end
end

function Npcm.printFramePos(frameName) -- pour dev
	local f = _G[frameName]
	if (Npcm.frameExists(f)) then
		print(frameName..":GetPoint()")
		local point, relativeTo, relativePoint, xOfs, yOfs = f:GetPoint()
		print("   point="..point)
		local relativeToName = "[UIParent]?" -- ?
		if (relativeTo ~= nil) then relativeToName = relativeTo:GetName(); end
		print("   relativeTo="..relativeToName)
		print("   relativePoint="..relativePoint)
		print("   xOfs="..floor(xOfs))
		print("   yOfs="..floor(yOfs))
	else
		print("Frame "..frameName.." introuvable.")
	end
end

---------------------------------------------------------------------------------------
------------------------------------ COMMUNICATION ------------------------------------

function Npcm.commPrefixOK(globalBoolean)
	if (globalBoolean == true) then
		return C_ChatInfo.RegisterAddonMessagePrefix(addonGChan)
	else
		return C_ChatInfo.RegisterAddonMessagePrefix(addonChan)
	end
end

------------------------------------------------- ENVOI

function Npcm.sendUpdateClient(i, comm) -- U : commande d'update de NPCM_SAVED clients par le MJ
	-- éventuellement limité au marker i, ou toutes les lignes si i vide
	-- success = C_ChatInfo.SendAddonMessage("prefix", "message" [, "chatType", "target"])
	local resu
	if (i == 0) then
		---------- update de tous les markers : nécessaire boucle car la limite d'envoi est de 250 caractères
		if (Npcm.updateMarkers()) then
			for c = 1, #markersOrder do
				resu = C_ChatInfo.SendAddonMessage(addonChan, Npcm.formatDataToSend(c, "U"), Npcm.getLocalChannel())
				if (not resu) then break; end
			end
			Npcm.testPrint(addonVar["name"].." updateClient() : ALL --> result="..tostring(resu))
		else
			Npcm.echoChatAppl(Npcm.getLocal("Unexpected markers error."), false, "red")
		end
	else
		---------- update du marker i
		resu = C_ChatInfo.SendAddonMessage(addonChan, Npcm.formatDataToSend(i, "U"), Npcm.getLocalChannel())
		Npcm.testPrint(addonVar["name"].." updateClient() : "..i.." --> result="..tostring(resu))
	end
	
	if (resu) then
		if (comm == nil) then
			if (i == 0) then comm = Npcm.getLocal("All players' markers"); else comm = Npcm.getLocal("Players' marker #")..i; end
		end
		Npcm.setComm(comm.." "..Npcm.getLocal("updated"), 10)
		resu = Npcm.sendPopulateUI() -- toutes les lignes de NPCM_SAVED sont d'abord mises à jour, puis affichées
	else
		Npcm.setComm(Npcm.getLocal("Unable to send data to players!"), 10)
	end
end

function Npcm.sendPopulateUI(showWarning) -- S ou W : envoi de la commande pour rafraichir les UI (MJ et client)
	Npcm.minimapOnOff(true) -- minimap (secu)
	if (Npcm.isTrue(NPCM_SAVED["settings"]["mj"]["flash"])) then Npcm.flashNPC() end
	local cmd = "S"
	if (showWarning) then cmd = "W" end
	local str = cmd..addonVar["sepData"]..Npcm.getPlayerFullName()..addonVar["sepData"]..Npcm.getVersionStr()
	return C_ChatInfo.SendAddonMessage(addonChan, str, Npcm.getLocalChannel())
end

function Npcm.sendVideClient() -- X : envoi par MJ commande de vidage au client
	local str = "X"..addonVar["sepData"]..Npcm.getPlayerFullName()..addonVar["sepData"]..Npcm.getVersionStr()
	C_ChatInfo.SendAddonMessage(addonChan, str, Npcm.getLocalChannel())
end

function Npcm.askUpdateRA() -- T : demande d'update du RL au RA
	local str = "T"..addonVar["sepData"]..Npcm.getPlayerFullName()..addonVar["sepData"]..Npcm.getVersionStr()
	if (not C_ChatInfo.SendAddonMessage(addonChan, str, Npcm.getLocalChannel())) then Npcm.echoChatAppl(Npcm.getLocal("Unable to update Raid Leader markers due to a communication prefix error."), false, "red") end
end

function Npcm.askUpdateClient() -- A : demande d'update au MJ de la part du client
	local str = "A"..addonVar["sepData"]..Npcm.getPlayerFullName()..addonVar["sepData"]..Npcm.getVersionStr()
	if (not C_ChatInfo.SendAddonMessage(addonChan, str, Npcm.getLocalChannel())) then Npcm.echoChatAppl(Npcm.getLocal("Unable to update client markers due to a communication prefix error."), false, "red") end
end

function Npcm.sendInactiveMJ() -- I : retour du MJ inactif : UI MJ non visible
	local str = "I"..addonVar["sepData"]..Npcm.getPlayerFullName()..addonVar["sepData"]..Npcm.getVersionStr()
	return C_ChatInfo.SendAddonMessage(addonChan, str, Npcm.getLocalChannel())
end

function Npcm.answerCheck(senderOr) -- R : réponse silencieuse à réception de C pour comptage log
	local str = "R"..addonVar["sepData"]..senderOr..addonVar["sepData"]..Npcm.booleanToInt(Npcm.isActivated())..addonVar["sepData"]..Npcm.getRank()..addonVar["sepData"]..Npcm.getVersionStr()
	return C_ChatInfo.SendAddonMessage(addonChan, str, Npcm.getLocalChannel())
end

function Npcm.sendCheck() -- C : check de communication sans envoi de sa version par commCheck()
	local str = "C"
	return C_ChatInfo.SendAddonMessage(addonChan, str, Npcm.getLocalChannel())
end

function Npcm.answerGlobalCheck(senderOr) -- K : réponse silencieuse à G (sur chan global : non limité au raid/party)
	local str = "K"..addonVar["sepData"]..senderOr..addonVar["sepData"]..Npcm.booleanToInt(Npcm.isActivated())..addonVar["sepData"]..Npcm.getStats()..addonVar["sepData"]..Npcm.getVersionStr()
	local chanID = Npcm.getGlobalChanID()
	if (chanID ~= 0) then
		return C_ChatInfo.SendAddonMessage(addonGChan, str, "CHANNEL", chanID)
		--return C_ChatInfo.SendAddonMessage(addonGChan, str, "CHANNEL", chanID.."."..addonVar["trpCompat"])
	else
		return false
	end
end

function Npcm.getGlobalChanID() -- renvoit 0 si channel introuvable
	return GetChannelName(addonVar["trpCompat"])
end

function Npcm.getStats() -- v >= 1.22
	local stats = 0
	if (Npcm.isMJ()) then stats = stats + 1 end
	if (NPCM_SAVED["settings"]["mj"]["darkui"] == 1) then stats = stats + 10 end
	stats = stats + (Npcm.strToNumber(NPCM_SAVED["settings"]["mj"]["hpprecision"]) / 100)
	return stats
end

function Npcm.videDataClient() -- vidage de NPCM_SAVED["markers"] du client et masquage frameClient
	---------- update de tous les markers : nécessaire boucle car la limite d'envoi est de 250 caractères
	wipe(NPCM_SAVED["markers"])
	for i = 1, #markersOrder do
		NPCM_SAVED["markers"][i] = {
			["text"] = "",
			["hp"] = 4,
			["boss"] = 0,
			["attitude"] = 0,
			["visible"] = 0,
		}
	end
	Npcm.testPrint(addonVar["name"].." videClient() : ALL")
	Npcm.activateClose()
end

function Npcm.formatDataToSend(i, cmd) -- formatage des données à envoyer (serialization)
	local sepData = addonVar["sepData"]
	local str = cmd..sepData..Npcm.getVersionStr()..sepData..tostring(i)..sepData..NPCM_SAVED["markers"][i]["visible"]..sepData..tostring(NPCM_SAVED["markers"][i]["hp"])..sepData..NPCM_SAVED["markers"][i]["boss"]..sepData..NPCM_SAVED["markers"][i]["attitude"]..sepData
	local desc = string.gsub(Npcm.videSiNul(NPCM_SAVED["markers"][i]["text"]), sepData, "==&==")
	str = str..strsub(desc, 1, 250 - strlen(str)) -- coupage de la desc pour ne pas dépasser un total de 250 caractères
	--[[
		1 - cmd (initiale)
		2 - addonVar["version"]
		3 - i
		4 - visible : 0-1
		5 - hp : 1-4
		6 - boss : 0-1
		7 - attitude : 0-2
		8 - desc
	--]]
	return str
end

function Npcm.getLocalChannel()
	if (Npcm.inRaid()) then
		return "RAID"
	elseif (Npcm.inInstance()) then
		return "INSTANCE_CHAT"
	else
		return "PARTY" -- default
	end
end

------------------------------------------------- RECEPTION

function Npcm.receivedMsg(prefix, text, channel, sender, target, zoneChannelID, localID, name, instanceID)
	-- gestion de l'event CHAT_MSG_ADDON (réception de message sur le canal de l'addon)
	if (prefix == addonChan) then
		local sepData = addonVar["sepData"]
		local playerName = Npcm.getPlayerFullName()
		if (sender ~= playerName) then
			local rank = Npcm.getRank()
			local cmd = strupper(strsub(text, 1, 1))
			--Npcm.testPrint("receivedMsg() : cmd = "..cmd.."-------- rank="..rank)) -- /////
			local tbl = {strsplit(sepData, text)}
			if (cmd == "U") then -------------- U : update data client (NPCM_SAVED) ou RA (ou RL si RA)
				if (#tbl >= 8) then
					local tableResu = {}
					tableResu["cmd"] = tbl[1]
					tableResu["vers"] = tbl[2]
					tableResu["index"] = tonumber(tbl[3])
					tableResu["visible"] = tbl[4]
					tableResu["hp"] = tonumber(tbl[5])
					tableResu["boss"] = tbl[6]
					tableResu["attitude"] = tbl[7]
					tableResu["text"] = string.gsub(tbl[8], "==&==", sepData)
					Npcm.updateLineData(tableResu)-- MAJ NPCM_SAVED d'après données reçues
				else
					return "Wrong parameters.   [821]."
				end
				
			elseif (cmd == "S") then -------------- S : commande de maj. des UI (MJ ou client) d'après NPCM_SAVED + version
				Npcm.showReceivedData(tbl)
				local mjVers = Npcm.getEnd(tbl)
				Npcm.checkReceivedVersion(mjVers, sender, false)
				
			elseif (cmd == "W") then -------------- W : comme S mais avec warning "assistant"
				Npcm.showReceivedData(tbl)
				local mjVers = Npcm.getEnd(tbl)
				Npcm.showHideAssistWarning(true)
				Npcm.setComm("You are now Assistant.", 10)
			
			elseif (cmd == "A") then -------------- A : demande du client vers MJ (seul le vrai MJ peut répondre)
				if (rank == 2) then -- vrai MJ (RL ou chef groupe)
					--Npcm.testPrint("receivedMsg() : demande d'update A si rank = 2 : "..Npcm.nilSiNul(tbl[2])) -- /////
					local clientVers = Npcm.getEnd(tbl)
					Npcm.checkReceivedVersion(clientVers, sender, false)
					if (npcm_frameMain:IsVisible()) then
						---------- UI MJ visible
						--Npcm.testPrint("receivedMsg() : demande d'update A de : "..Npcm.nilSiNul(tbl[2])) -- /////
						Npcm.sendUpdateClient(0, nil) -- envoit de U (toutes lignes) et S pour update et affichage (clients et RA)
					else
						---------- UI MJ non visible (MJ inactif)
						Npcm.sendInactiveMJ() -- renvoit de I : "mon UI est pas visible : masquer toutes les UI"
					end
				end
				
			elseif (cmd == "T") then ------------- T : demande du nouveau RL vers RA (seul le RA peut répondre)
				if (rank == 1) then -- RA
					if (Npcm.debug()) then print("receivedMsg() : demande d'update T au RA de : "..Npcm.nilSiNul(tbl[2])) end
					if (npcm_frameMain:IsVisible()) then
						---------- UI MJ visible
						-- Npcm.testPrint("receivedMsg() : demande d'update T au RA de : "..Npcm.nilSiNul(tbl[2])) -- /////
						Npcm.sendUpdateClient(0, nil) -- envoir de U (toutes lignes) et S pour update et affichage (clients et RA)
					else
						---------- UI MJ non visible (MJ inactif)
						Npcm.sendInactiveMJ() -- renvoit de I : "mon UI est pas visible : masquer toutes les UI"
					end
				end
				
			elseif (cmd == "I") then ------------- I : retour de demande d'update au MJ : le vrai MJ est inactif
				if (rank < 2) then -- tous clients et RA
					--Npcm.testPrint("receivedMsg() : retour inactif I de : "..Npcm.nilSiNul(tbl[2])) -- /////
					Npcm.activateClose()
				end
			
			elseif (cmd == "X") then ------------- X : commande de vidage des data client et masquage frame client : uniquement
				if (rank < 2) then Npcm.videDataClient() end -- tous clients et RA
				
			elseif (cmd == "C") then ------------- C : check
				Npcm.answerCheck(sender) -- retour : R envoyé
				
			elseif (cmd == "R") then ------------- R : retour d'un check pour comptage log
				-- retour v < 1.03 : "R"..addonVar["sepData"]..senderOr..addonVar["sepData"]..Npcm.getVersionStr()
				local clientVers = Npcm.getEnd(tbl)
				Npcm.checkReceivedVersion(clientVers, sender, false)
				local senderOr = Npcm.videSiNul(tbl[2])
				local active, rank
				if (tbl[4] ~= nil) then -- v 1.03
					active = tonumber(Npcm.zeroSiNul(tbl[3], true))
					rank = tonumber(Npcm.zeroSiNul(tbl[3], true))
				end
				Npcm.testPrint("R received from "..sender.." (senderOr="..senderOr..")")
				if (senderOr == playerName) then -- c'est bien playerName qui a envoyé la requete "C"
					Npcm.logMembers(sender, clientVers, active, rank) -- compter réponses avec timer
				end
				
			end
			
			--if (Npcm.debug()) then print("receivedMsg()---------", "prefix="..tostring(prefix), "text="..tostring(text), "channel="..tostring(channel), "sender="..tostring(sender),"target="..tostring(target), "zoneChannelID="..tostring(zoneChannelID), "localID="..tostring(localID), "name="..tostring(name), "instanceID="..tostring(instanceID)) end
			return true
		end
		return true
	
	elseif (prefix == addonGChan) then
		local playerName = Npcm.getPlayerFullName()
		if (sender ~= playerName) then
			local sepData = addonVar["sepData"]
			local cmd = strupper(strsub(text, 1, 1))
			local tbl = {strsplit(sepData, text)}
			
			if (cmd == "G") then -------------------- G : ping check global
				local vers = Npcm.strToNumber(Npcm.getEnd(tbl))
				Npcm.testPrint("G received from "..sender.." with version "..vers)
				if (vers > 0) then Npcm.checkReceivedVersion(vers, sender, true) end -- forcage de vérification de version si version reçue > 0
				Npcm.answerGlobalCheck(sender) -- retour K sur le chan global
			end
		end
	else
		--if (Npcm.debug()) then print("receivedMsg()---------", "prefix="..tostring(prefix), "channel="..tostring(channel), "sender="..tostring(sender),"target="..tostring(target), "zoneChannelID="..tostring(zoneChannelID), "localID="..tostring(localID), "name="..tostring(name), "text="..tostring(text)) end
	end
	return false
end

function Npcm.updateLineData(tableLigne) -- MAJ NPCM_SAVED d'après données reçues
	if (type(tableLigne) == "table") then -- secu
		local index = tableLigne["index"]
		------------ sauvegarde pour comparaison ultérieure
		prevSavedMarkers[index]["visible"] = NPCM_SAVED["markers"][index]["visible"]
		prevSavedMarkers[index]["attitude"] = NPCM_SAVED["markers"][index]["attitude"]
		prevSavedMarkers[index]["hp"] = NPCM_SAVED["markers"][index]["hp"]
		------------ 
		NPCM_SAVED["markers"][index] = {
			["text"] = tableLigne["text"],
			["hp"] = tableLigne["hp"],
			["boss"] = tableLigne["boss"],
			["attitude"] = tableLigne["attitude"],
			["visible"] = tableLigne["visible"],
		}
	end
end

function Npcm.showReceivedData(tbl) -- MAJ des UI (client et MJ) d'après NPCM_SAVED
	--------------------------- MAJ des frames MJ si RA, sinon du client
	local rank = Npcm.getRank()
			
	if (rank >= 1) then -- RL, RA ou chef groupe
		npcm_frameClient:Hide() -- secu
		Npcm.populateFrameMJ() -- ne pas effacer le comm existant
		npcm_frameMain:Show()
		Npcm.updatePartyIcon(rank)
	elseif (rank == 0) then -- joueur
		npcm_frameMain:Hide() -- secu
		Npcm.compareClientMarkers()
		Npcm.populateFrameClient()
	end
	Npcm.flashNPC()
	Npcm.minimapOnOff(true) -- minimap
end

function Npcm.compareClientMarkers() -- comparaison ancien markers et SAVED pour jouage son
	local totalPrevVisible, totalPrevAttitude, totalPrevHP = Npcm.getMarkersMap(prevSavedMarkers, true)
	local totalSavedVisible, totalSavedAttitude, totalSavedHP = Npcm.getMarkersMap(NPCM_SAVED["markers"], false)
	
	if (totalSavedVisible > totalPrevVisible) then
		Npcm.testPrint("PLAYAGE : place -----")
		Npcm.playUIsounds("client", "place")
	elseif (totalSavedVisible < totalPrevVisible) then
		Npcm.testPrint("PLAYAGE : clear -----")
		Npcm.playUIsounds("client", "clear")
	else
		if (totalSavedVisible > 0) then
		--print("------ totalPrevAttitude="..totalPrevAttitude.." --> totalSavedAttitude="..totalSavedAttitude.."   totalPrevHP="..totalPrevHP.." -->totalSavedHP="..totalSavedHP)
			if (totalSavedAttitude ~= totalPrevAttitude) then
				Npcm.testPrint("PLAYAGE : attitude -----")
				Npcm.playUIsounds("client", "attitude")
			elseif (totalSavedHP ~= totalPrevHP) then
				Npcm.testPrint("PLAYAGE : hp -----")
				Npcm.playUIsounds("client", "hp")
			end
		end
	end
	-- NOBOSS
	
	-- prevSavedMarkers = NPCM_SAVED["markers"] -- NON avec ce con de lua !!!
	for i = 1, #markersOrder do
		for key, value in pairs(NPCM_SAVED["markers"][i]) do
			prevSavedMarkers[i][key] = value
		end
	end
end

function Npcm.getMarkersMap(markersTbl, devPrint)
	local totalVisible, totalAttitude, totalHP = 0, 0, 0
	for i, value in pairs(markersTbl) do
		if (tonumber(value["visible"]) == 1) then totalVisible = totalVisible + 1 end
		if (tonumber(value["attitude"]) == 1) then totalAttitude = totalAttitude + 1 end
		totalHP = totalHP + tonumber(value["hp"])
	end
	return totalVisible, totalAttitude, totalHP
	-- NOBOSS
end

------------------------------------------ VERSION

function Npcm.getVersionStr() -- raccourci (renvoit une chaine)
	return (Npcm.getVar("version")..Npcm.videSiNul(Npcm.getVar("publicVersion")))
end

function Npcm.checkReceivedVersion(versReceived, sender, forceBoolean)
	if (Npcm.strToNumber(versReceived) > 0) then -- comparaison seulement si versReceived est public
		if (forceBoolean == true) then NPCM_SAVED["settings"]["dayVerified"] = 0 end
		if ((Npcm.getDaysEpoch() - NPCM_SAVED["settings"]["dayVerified"]) > numDaysCheckVers) then
			Npcm.testPrint("checkReceivedVersion("..Npcm.nilSiNul(sender)..") : Received="..tonumber(versReceived).." Addon compat="..addonVar["compatVersion"].." Addon vers="..tonumber(addonVar["version"]))
			if (tonumber(versReceived) > tonumber(addonVar["version"])) then
				local str = Npcm.getLocal("Your version is outdated").." ("..Npcm.getVersionStr().."). "..Npcm.getLocal("New version available:")..versReceived
				Npcm.echoChatAppl(": ".. str, true, "red")
			end
			NPCM_SAVED["settings"]["dayVerified"] = Npcm.getDaysEpoch()
		end
	end
end

function Npcm.getDaysEpoch()
	return floor(time() / 60 / 60 / 24)
end

---------------------------------------------------------------------------------------
---------------------------------------- MINIMAP --------------------------------------

function Npcm.minimapIconShow() -- affichage du bouton minimap
	NPCM_SAVED["settings"]["minimap"]["visible"] = 1
	_G[addonVar["minimap"]]:Show()
end

function Npcm.minimapOnOff(booleen)
	local iconTexture = _G[addonVar["minimap"].."_tex"]
	if (Npcm.isTrue(booleen)) then
		iconTexture:SetTexture(Npcm.getStuffPath("icnMinimap", true)) -- actif
	else
		iconTexture:SetTexture("Interface\\TargetingFrame\\UI-RaidTargetingIcon_1") -- inactif (ou 5 pour lune)
	end
end

---------------------------------------------------------------------------------------
------------------------------------- LIB DEDIEE --------------------------------------

function Npcm.debug()
	return Npcm.isTrue(NPCM_SAVED["settings"]["debug"])
end

function Npcm.getVar(nomVar, subLvl1, subLvl2)
	-- get valeurs de la table locale addonVar
	-- si aucun argument n'est passé, renvoit la table addonVar
	if (addonVar == nil) then return nil end
	
	if (subLvl2 ~= nil) then
		if (addonVar[nomVar][subLvl1][subLvl2] ~= nil) then return addonVar[nomVar][subLvl1][subLvl2] end
	elseif (subLvl1 ~= nil) then
		if (addonVar[nomVar][subLvl1] ~= nil) then return addonVar[nomVar][subLvl1] end
	elseif (nomVar ~= nil) then
		if (addonVar[nomVar] ~= nil) then return addonVar[nomVar] end
	else
		return addonVar
	end
end

function Npcm.setVar(nomVar, valeur) -- nécessite addonVar
	if (addonVar == nil) then return nil end
	if (nomVar ~= nil) then
		if (addonVar[nomVar] ~= nil) then addonVar[nomVar] = valeur end
	end
end

function Npcm.getSavedSettings(perCharacterBoolean) -- pour appel depuis autres fichiers
	if (perCharacterBoolean) then return NPCM_SAVEDPERCHAR else return NPCM_SAVED end
end
