Npcm = {} -- prefixe fonctions
local addonName = ...

local compatVersion = tonumber(GetAddOnMetadata(addonName, "X-CompatibleVersion")) -- /////////// mettre à jour la TOC
local publicVersion = strlower(GetAddOnMetadata(addonName, "X-PublicVersion")) -- /////////// mettre à jour la TOC : "public" ou "" | "alpha" | "beta"

local addonVar = {
	["name"] = addonName,
	["version"] = GetAddOnMetadata(addonName, "Version"),
	["compatVersion"] = compatVersion,
	["publicVersion"] = publicVersion,
	["trpCompat"] = "xtensionxtooltip2", -- mettre à vide pour ne pas utiliser ce chan utilisé par TRP
	["sepData"] = "&", -- communication
	["chatColor"] = "AE7115",
	["slashCmd"] = "npcsmarkers", -- nécessaire
	["settingsPanel"] = "npcmCfgFrame", -- nom du frame de settings : laisser vide si utilisation du frame de WoW
	["minimap"] = "Npcm_minimapBtn", -- nécessite le fichier _minimap.lua (laisser vide ou NIL si pas de bouton minimap)
	["dimUImj"] = { -- params de la scrolling raid/party players list : voir Npcm.listGetVar() du fichier UI
		["hFrame"] = 280, -- largeur du main frame
		["vLine"] = 60, -- hauteur d'une "ligne" de marker
		["vMenuBar"] = 16, -- hauteur de la barre de titre
		["vStatus"] = 16, -- hauteur de la barre de status bas
	},
	["dimUIclient"] = {
		["hFrame"] = 130, -- largeur du main frame
		["vLine"] = 32, -- hauteur d'une "ligne" de marker
		["icon"] = 22, -- taille d'un marker
	},
}

function Npcm.initData(subLvl1, subLvl2) -- subLvl1, subLvl2 : args optionnels pour les "settings"
	local tbl = {
		["settings"] = {
			["version"] = addonVar["version"],
			["dayVerified"] = 0,
			["debug"] = 0,
			["forceClient"] = 0,
			["visible"] = 1,
			["minimap"] = {
				["visible"] = 1,
			},
			["mj"] = {
				["expanded"] = 1,
				["alpha"] = 1,
				["darkalpha"] = 1,
				["posX"] = 100,
				["posY"] = 700, -- distance bas UIParent / haut main frame
				["scale"] = 1,
				["showWarning"] = 1,
				["flash"] = 1,
				["sounds"] = 1,
				["hpprecision"] = 4,
				["darkui"] = 1,
			},
			["client"] = {
				["alpha"] = 0.8,
				["posX"] = 70,
				["posY"] = 600, -- distance bas UIParent / haut main frame
				["scale"] = 1,
				["sounds"] = 1,
			},
		},
		["markers"] = Npcm.initSavedMarkers(),
	}
	
	if (subLvl2 ~= nil) then
		if (tbl["settings"][subLvl1][subLvl2] ~= nil) then return tbl["settings"][subLvl1][subLvl2] end
	elseif (subLvl1 ~= nil) then
		if (tbl["settings"][subLvl1] ~= nil) then return tbl["settings"][subLvl1] end
	else
		return tbl
	end
end

function Npcm.initSavedMarkers()
	local tbl = {}
	for i = 1, #Npcm.getMarkersOrder() do
		tbl[i] = Npcm.getEmptyMarkerData()
	end
	return tbl
end

function Npcm.getEmptyMarkerData()
	return ({
		["text"] = "",
		["hp"] = 100, -- %
		["boss"] = 0,
		["attitude"] = 3, -- gris
		["visible"] = 0,
	})
end

local markersOrder = { -- pour remplacer WORLD_RAID_MARKER_ORDER qui semble parfois générer une erreur lua
	[1] = 8,
	[2] = 4,
	[3] = 1,
	[4] = 7,
	[5] = 2,
	[6] = 3,
	[7] = 6,
	[8] = 5,
}
function Npcm.getMarkersOrder() -- pour autre fichier
	return markersOrder
end

function Npcm.getSounds(action)
	local sounds = {
		["hp"] = {["id"] = 835, ["name"] = "igAbilityClose"},
		["attitude"] = {["id"] = 1221, ["name"] = "PickupGems"},
		["name"] = {["id"] = 1193, ["name"] = "PickUpRing"},
		["place3"] = {["id"] = 2560, ["name"] = "SPELL_Shadow_Cast"},
		["clear"] = {["id"] = 47747, ["name"] = "Garrisons_HolyPaladin_Cast"}, -- clear marker
		["place"] = {["id"] = 39, ["name"] = "Divine Shield"}, -- place marker
		["ui"] = {["id"] = 857, ["name"] = "igMainMenuOptionCheckBoxOff"}, -- action UI MJ sur ligne non visible
	}
	-- 684 -- INTERFACESOUND_LOSTTARGETUNIT
	return sounds[action]["id"]
end

---------------------------------------------------------------------------------------
------------------------------------- LIB DEDIEE --------------------------------------

function Npcm.debug()
	return Npcm.isTrue(NPCM_SAVED["settings"]["debug"])
end

function Npcm.getVar(nomVar, subLvl1, subLvl2)
	-- get valeurs de la table locale addonVar
	-- si aucun argument n'est passé, renvoit la table addonVar
	if (addonVar == nil) then return nil end
	
	if (subLvl2 ~= nil) then
		if (addonVar[nomVar][subLvl1][subLvl2] ~= nil) then return addonVar[nomVar][subLvl1][subLvl2] end
	elseif (subLvl1 ~= nil) then
		if (addonVar[nomVar][subLvl1] ~= nil) then return addonVar[nomVar][subLvl1] end
	elseif (nomVar ~= nil) then
		if (addonVar[nomVar] ~= nil) then return addonVar[nomVar] end
	else
		return addonVar
	end
end

function Npcm.setVar(nomVar, valeur) -- nécessite addonVar
	if (addonVar == nil) then return nil end
	if (nomVar ~= nil) then
		if (addonVar[nomVar] ~= nil) then addonVar[nomVar] = valeur end
	end
end

function Npcm.getSavedSettings(perCharacterBoolean) -- pour appel depuis autres fichiers
	if (perCharacterBoolean) then return NPCM_SAVEDPERCHAR else return NPCM_SAVED end
end
